(*
    Id3Tag - Main form
    Copyright (c) 1998-2003,2004 Alessandro Scotti
    http://www.ascotti.org/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)
unit Main;

interface

uses
  Id3,
  SysUtils, Classes, QForms, QImgList, QDialogs, QClipBrd,
  QMenus, QTypes, QComCtrls, QControls, QExtCtrls;

type
  TSortMode = ( byNone, byFilename, byTitle, byArtist, byAlbum, byTrack, byYear, byGenre, byComment );

type
  TformMain = class(TForm)
    mnuMain: TMainMenu;
    mnuFile: TMenuItem;
    barStatus: TStatusBar;
    dlgOpen: TOpenDialog;
    lstFiles: TListView;
    cmdOpen: TMenuItem;
    cmdN1: TMenuItem;
    cmdExit: TMenuItem;
    popupFiles: TPopupMenu;
    cmdProperties: TMenuItem;
    mnuHelp: TMenuItem;
    cmdAbout: TMenuItem;
    cmdSave: TMenuItem;
    mnuEdit: TMenuItem;
    cmdSelectAll: TMenuItem;
    cmdCut: TMenuItem;
    cmdCopy: TMenuItem;
    cmdPaste: TMenuItem;
    cmdN3: TMenuItem;
    cmdEditProperties: TMenuItem;
    mnuTools: TMenuItem;
    cmdGuessTrackNumber: TMenuItem;
    cmdMakeHTMLIndex: TMenuItem;
    dlgSavePlaylist: TSaveDialog;
    N1: TMenuItem;
    cmdSavePlaylist: TMenuItem;
    cmdRename: TMenuItem;
    cmdCapitalizeTitle: TMenuItem;
    barMain: TToolBar;
    tbbNew: TToolButton;
    imlToolbar: TImageList;
    tbbOpen: TToolButton;
    tbbSave: TToolButton;
    tbbSep1: TToolButton;
    tbbCopy: TToolButton;
    tbbCut: TToolButton;
    tbbPaste: TToolButton;
    tbbSep2: TToolButton;
    tbbEdit: TToolButton;
    dlgSaveIndex: TSaveDialog;
    N2: TMenuItem;
    cmdNew: TMenuItem;
    procedure lstFilesCompare(Sender: TObject; Item1, Item2: TListItem;
      Data: Integer; var Compare: Integer);
    procedure FormCreate(Sender: TObject);
    procedure lstFilesColumnClick(Sender: TObject; Column: TListColumn);
    procedure cmdOpenClick(Sender: TObject);
    procedure lstFilesDblClick(Sender: TObject);
    procedure lstFilesMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure cmdPropertiesClick(Sender: TObject);
    procedure cmdRenameClick(Sender: TObject);
    procedure cmdSaveClick(Sender: TObject);
    procedure cmdPlayClick(Sender: TObject);
    procedure cmdSelectAllClick(Sender: TObject);
    procedure cmdAboutClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure cmdExitClick(Sender: TObject);
    procedure lstFilesInsert(Sender: TObject; Item: TListItem);
    procedure lstFilesDeletion(Sender: TObject; Item: TListItem);
    procedure lstFilesKeyPress(Sender: TObject; var Key: Char);
    procedure cmdCutClick(Sender: TObject);
    procedure cmdCopyClick(Sender: TObject);
    procedure mnuEditClick(Sender: TObject);
    procedure cmdGuessTrackNumberClick(Sender: TObject);
    procedure cmdSavePlaylistClick(Sender: TObject);
    procedure cmdMakeHTMLIndexClick(Sender: TObject);
    procedure cmdCapitalizeTitleClick(Sender: TObject);
    procedure tbbNewClick(Sender: TObject);
    procedure lstFilesEdited(Sender: TObject; Item: TListItem;
      var S: WideString);
  private
    { Private declarations }
    SortMode: TSortMode;
    SortInverse: boolean;
    procedure addId3ToList( const Filename: string; Id3: TId3Info );
    procedure setId3OfItem( Item: TListItem );
    procedure AddSingleFile( Name: string );
    procedure AddFile( const Name: string );
    procedure ToClipboard( Cut: boolean );
    procedure PopulatePopupMenu;
    function AutoSave( Item: TListItem; var Res: integer ): integer;
    function NextListItem( Item: TListItem; States: TItemStates ): TListItem;
    function VerifySaved( SelectedOnly: boolean ): boolean;
  public
    procedure Apply( Info: TId3Info );
  end;

var
  formMain: TformMain;

implementation

uses
  Types, Properties, Rename, About, ConfirmSave,
  MakeIndex, Tools, SaveFiles, IniFiles;

{$R *.dfm}

procedure TformMain.setId3OfItem( Item: TListItem );
var
  Id3: TId3Info;
begin
  Id3 := TId3Info( Item.Data );
  Item.Subitems.Clear;
  Item.Subitems.Add( Id3.Title );
  Item.Subitems.Add( Id3.Artist );
  Item.Subitems.Add( Id3.Album );
  if( Id3.Track = 0 ) then
    Item.Subitems.Add( '' )
  else
    Item.Subitems.Add( Format( '%.2d', [Id3.Track] ) );
  Item.Subitems.Add( Id3.Year );
  Item.Subitems.Add( GetGenreName( Id3.Genre ) );
  Item.Subitems.Add( Id3.Comment );
end;

procedure TformMain.addId3ToList( const Filename: string; Id3: TId3Info );
var
  Item: TListItem;
begin
  Item := lstFiles.Items.Add;
  Item.Caption := Filename;
  Item.Data := Id3;
  Item.Selected := true;
  setId3OfItem( Item );
end;

function Sign( N: integer ): integer;
begin
  Sign := -1;
  if( N = 0 ) then
    Sign := 0
  else if( N > 0 ) then
    Sign := 1;
end;

procedure TformMain.lstFilesCompare(Sender: TObject; Item1,
  Item2: TListItem; Data: Integer; var Compare: Integer);
var
  Inf1, Inf2: TId3Info;
begin
  Inf1 := TId3Info( Item1.Data );
  Inf2 := TId3Info( Item2.Data );
  case SortMode of
  byNone,
  byFilename: Compare := AnsiCompareStr( Item1.Caption, Item2.Caption );
  byTitle: Compare := AnsiCompareStr( Inf1.Title, Inf2.Title );
  byArtist: Compare := AnsiCompareStr( Inf1.Artist, Inf2.Artist );
  byAlbum: Compare := AnsiCompareStr( Inf1.Album, Inf2.Album );
  byYear: Compare := AnsiCompareStr( Inf1.Year, Inf2.Year );
  byGenre: Compare := AnsiCompareStr( GetGenreName(Inf1.Genre), GetGenreName(Inf2.Genre) );
  byComment: Compare := AnsiCompareStr( Inf1.Comment, Inf2.Comment );
  byTrack: Compare := Sign( Inf1.Track - Inf2.Track );
  end;
  if( SortInverse )and( Compare <> 0 ) then begin
    Compare := -Compare;
  end;
end;

procedure TformMain.lstFilesColumnClick(Sender: TObject;
  Column: TListColumn);
begin
  lstFiles.Sort( Column.Index, sdAscending );
end;

procedure TformMain.cmdOpenClick(Sender: TObject);
var
  Idx: integer;
begin
  if( dlgOpen.Execute ) then begin
    try
      lstFiles.Items.BeginUpdate;
      Screen.Cursor := crHourGlass;
      for Idx := 0 to dlgOpen.Files.Count-1 do begin
        AddSingleFile( dlgOpen.Files[Idx] );
      end;
      if( dlgOpen.Files.Count > 0 ) then
        dlgOpen.InitialDir := ExtractFilePath( dlgOpen.Files[0] );
    finally
      Screen.Cursor := crDefault;
      lstFiles.Items.EndUpdate;
    end;
  end;
end;

procedure TformMain.lstFilesDblClick(Sender: TObject);
begin
  cmdPropertiesClick( Sender );
end;

procedure TformMain.lstFilesMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  P: TPoint;
begin
  if( lstFiles.SelCount > 0 )and( Button = mbRight ) then begin
    P := lstFiles.ClientToScreen( Point( X, Y ) );
    popupFiles.Popup( P.X, P.Y );
  end;
end;

procedure TformMain.cmdPropertiesClick(Sender: TObject);
var
  Item: TListItem;
  Inf1: TId3Info;
  Inf2: TId3Info;
begin
  Item := lstFiles.Selected;
  if( Item = nil ) then
    Exit;
  if( lstFiles.SelCount = 1 ) then begin
    formId3Properties.Filename := Item.Caption;
    formId3Properties.Info := TId3Info( Item.Data );
  end
  else begin
    formId3Properties.Filename := '';
    Inf1 := TId3Info.Create;
    Inf1.Assign( TId3Info(Item.Data) );
    (* Scan the list of selected items and add the tag information *)
    (* to the current tag: this has the effect of setting the proper *)
    (* "changed" flag for each field *)
    repeat
      Item := NextListItem( Item, [isSelected] );
      if( Item <> nil ) then begin
        Inf2 := TId3Info( Item.Data );
        Inf1.Title := Inf2.Title;
        Inf1.Album := Inf2.Album;
        Inf1.Artist := Inf2.Artist;
        Inf1.Genre := Inf2.Genre;
        Inf1.Year := Inf2.Year;
        Inf1.Comment := Inf2.Comment;
      end;
    until( Item = nil );
    (* Now the "changed" fields are those which are not the same *)
    (* across all selected entries, so clear them as there is no *)
    (* meaningful default *)
    if( ifTitleChanged in Inf1.Flags ) then
      Inf1.Title := '';
    if( ifArtistChanged in Inf1.Flags ) then
      Inf1.Artist := '';
    if( ifAlbumChanged in Inf1.Flags ) then
      Inf1.Album := '';
    if( ifYearChanged in Inf1.Flags ) then
      Inf1.Year := '';
    if( ifGenreChanged in Inf1.Flags ) then
      Inf1.Genre := -1;
    if( ifCommentChanged in Inf1.Flags ) then
      Inf1.Comment:= '';
    (* Reset the "changed" flags before editing *)
    Inf1.ClearChanged;
    formId3Properties.Info := Inf1;
    Inf1.Free;
  end;
  if( formId3Properties.ShowModal = mrOk ) then
    Apply( formId3Properties.Info );
  // Reset info in case the same item is edited again right now
  formId3Properties.Info := nil;
end;

procedure TformMain.Apply( Info: TId3Info );
var
  Item: TListItem;
  ItemInfo: TId3Info;
begin
  Item := lstFiles.Selected;
  while( Item <> nil ) do begin
    ItemInfo := TId3Info( Item.Data );
    if( ifTitleChanged in Info.Flags ) then
      ItemInfo.Title := Info.Title;
    if( ifArtistChanged in Info.Flags ) then
      ItemInfo.Artist:= Info.Artist;
    if( ifAlbumChanged in Info.Flags ) then
      ItemInfo.Album := Info.Album;
    if( ifYearChanged in Info.Flags ) then
      ItemInfo.Year := Info.Year;
    if( ifGenreChanged in Info.Flags ) then
      ItemInfo.Genre := Info.Genre;
    if( ifCommentChanged in Info.Flags ) then
      ItemInfo.Comment:= Info.Comment;
    if( ifTrackChanged in Info.Flags ) then
      ItemInfo.Track:= Info.Track;
    setId3OfItem( Item );
    Item := NextListItem( Item, [isSelected] );
  end;
end;

procedure TformMain.cmdRenameClick(Sender: TObject);
var
  Item: TListItem;
  INew: TListItem;
  Name: string;
  I: integer;
begin
  Item := lstFiles.Selected;
  if( Item = nil ) then
    Exit;
  formRename.lstFiles.Items.Clear;
  for I:=1 to lstFiles.SelCount do begin
    Assert( Item <> nil );
    INew := formRename.lstFiles.Items.Add;
    INew.Caption := ChangeFileExt( ExtractFileName(Item.Caption), '' );
    INew.Data := Item.Data;
    Item := NextListItem( Item, [isSelected] );
  end;
  if( formRename.ShowModal <> mrOk ) then
    Exit;
  (* Rename *)
  Item := lstFiles.Selected;
  for I:=1 to lstFiles.SelCount do begin
    Assert( Item <> nil );
    Name := ChangeFileExt( ExtractFileName(Item.Caption), '' );
    Name := formRename.GetNameFromId3( formRename.GetNameFormat, TId3Info(Item.Data) );
    if( Name <> '' )and( Name[1] <> ' ' ) then begin
      Name := ExtractFilePath(Item.Caption) + Name + ExtractFileExt(Item.Caption);
      if( RenameFile( Item.Caption, Name ) ) then begin
        Item.Caption := Name;
      end;
    end;
    Item := NextListItem( Item, [isSelected] );
  end;
end;

procedure TformMain.AddSingleFile( Name: string );
var
  Info: TId3Info;
begin
  Info := TId3Info.Create;
  Info.LoadFromFile( Name );
  addId3ToList( Name, Info );
end;

procedure TformMain.AddFile( const Name: string );
var
  Path: string;
  SR: TSearchRec;
begin
  Path := Name;
  if( DirectoryExists(Path) ) then
    Path := Path + '/*';
  if( Pos('*',Path) > 0 )or( Pos('?',Path) > 0 ) then begin
    if( FindFirst( Path, faAnyFile, SR ) = 0 ) then begin
      repeat
        if( (SR.Attr and (faDirectory or faVolumeID or faSysFile)) = 0 ) then
          AddSingleFile( ExtractFilePath(Path) + SR.Name );
      until( FindNext( SR ) <> 0 );
      FindClose( SR );
    end;
  end
  else begin
    AddSingleFile( Path );
  end;
end;

procedure TformMain.cmdSaveClick(Sender: TObject);
var
  Idx: integer;
  Inf: TId3Info;
begin
  for Idx:=0 to lstFiles.Items.Count-1 do begin
    Inf := TId3Info( lstFiles.Items[Idx].Data );
    if( Inf.Changed ) then begin
      Inf.SaveToFile( lstFiles.Items[Idx].Caption );
      Inf.ClearChanged;
    end;
  end;
end;

procedure TformMain.cmdPlayClick(Sender: TObject);
begin
  // Nothing in the Linux version...
end;

procedure TformMain.cmdSelectAllClick(Sender: TObject);
var
  I: integer;
begin
  lstFiles.Items.BeginUpdate;
  for I:=0 to lstFiles.Items.Count-1 do
    lstFiles.Items[I].Selected := true;
  lstFiles.Items.EndUpdate;
end;

procedure TformMain.cmdAboutClick(Sender: TObject);
begin
  formAbout := TFormAbout.Create( Self );
  formAbout.ShowModal;
  formAbout.Free;
end;

const
  sOptions = 'Options';

  sWinMax = 'Maximized';
  sWinLeft = 'Left';
  sWinTop = 'Top';
  sWinWidth = 'Width';
  sWinHeight = 'Height';
  sLstColumn = 'ColWidth';
  sSortMode = 'SortMode';
  sSortInverse = 'SortInverse';
  sInitialDir = 'InitialDir';

procedure TformMain.FormCreate(Sender: TObject);
var
  Idx: integer;
  Ini: TIniFile;
begin
  SortMode := byNone;
  SortInverse := false;

  (* Add files if specified from command line *)
  for Idx:=1 to ParamCount do begin
    AddFile( ParamStr(Idx) );
  end;

  (* *)
  Ini := TIniFile.Create( ChangeFileExt(Application.ExeName,'.ini') );
  try
    (* Read the window state *)
    if( Ini.ReadBool( sOptions, sWinMax, false ) ) then begin
      WindowState := wsMaximized
    end
    else begin
      Left := Ini.ReadInteger( sOptions, sWinLeft, Left );
      Top := Ini.ReadInteger( sOptions, sWinTop, Top );
      Width := Ini.ReadInteger( sOptions, sWinWidth, Width );
      Height := Ini.ReadInteger( sOptions, sWinHeight, Height );
    end;

    (* Read the columns width *)
    for Idx:=0 to lstFiles.Columns.Count-1 do begin
      lstFiles.Columns[Idx].Width := Ini.ReadInteger( sOptions, sLstColumn + IntToStr(Idx), lstFiles.Columns[Idx].Width );
    end;

    (* Read the sort mode *)
    SortMode := TSortMode( Ini.ReadInteger( sOptions, sSortMode, 0 ) );
    SortInverse := Ini.ReadBool( sOptions, sSortInverse, false );

    (* Read the default directory *)
    dlgOpen.InitialDir := Ini.ReadString( sOptions, sInitialDir, '' );
  finally
    Ini.Free;
  end;

  (* Populate right-click popup menu *)
  PopulatePopupMenu;
end;

procedure TformMain.PopulatePopupMenu;
begin
  // Nothing in the Linux version...
end;

function TformMain.VerifySaved( SelectedOnly: boolean ): boolean;
var
  Idx : integer;
  Res : integer;
  Info: TId3Info;
  Item: TListItem;
begin
  Result := false;

{$ifdef VERSION_1_01}
  Res := mrNone;
  for Idx:=0 to lstFiles.Items.Count-1 do begin
    AutoSave( lstFiles.Items[Idx], Res );
    if( Res = mrCancel ) then
      Exit;
  end;
{$else}
  (* Save files that are changed *)
  formSaveFiles.lstMain.Items.BeginUpdate;
  formSaveFiles.lstMain.Items.Clear;
  for Idx:=0 to lstFiles.Items.Count-1 do begin
    Info := TId3Info(lstFiles.Items[Idx].Data);
    if Info.Changed then begin
      Item := formSaveFiles.lstMain.Items.Add;
      Item.Caption := lstFiles.Items[Idx].Caption;
      Item.Data := lstFiles.Items[Idx].Data;
      Item.Selected := true;
    end;
  end;
  formSaveFiles.lstMain.Items.EndUpdate;
  if( formSaveFiles.lstMain.Items.Count > 0 ) then begin
    Res := formSaveFiles.ShowModal;
    if( Res = mrCancel ) then
      Exit;
    if( Res = mrYes ) then begin
      for Idx:=0 to formSaveFiles.lstMain.Items.Count-1 do begin
        Item := formSaveFiles.lstMain.Items[Idx];
        Info := TId3Info(Item.Data);
        Info.SaveToFile( Item.Caption );
      end;
    end;
  end;
{$endif}

  Result := true;
end;

procedure TformMain.FormClose(Sender: TObject; var Action: TCloseAction);
var
  Idx : integer;
  Ini: TIniFile;
begin
  if( not VerifySaved( false ) ) then begin
    Action := caNone;
    Exit;
  end;

  (* *)
  Ini := TIniFile.Create( ChangeFileExt(Application.ExeName,'.ini') );
  try
    (* Save the window state *)
    Ini.WriteBool( sOptions, sWinMax, WindowState = wsMaximized );
    Ini.WriteInteger( sOptions, sWinLeft, Left );
    Ini.WriteInteger( sOptions, sWinTop, Top );
    Ini.WriteInteger( sOptions, sWinWidth, Width );
    Ini.WriteInteger( sOptions, sWinHeight, Height );
    (* Save the columns width *)
    for Idx:=0 to lstFiles.Columns.Count-1 do begin
      Ini.WriteInteger( sOptions, sLstColumn + IntToStr(Idx), lstFiles.Columns[Idx].Width );
    end;
    (* Save the sort mode *)
    Ini.WriteInteger( sOptions, sSortMode, integer(SortMode) );
    Ini.WriteBool( sOptions, sSortInverse, SortInverse );
    (* Write the default directory *)
    Ini.WriteString( sOptions, sInitialDir, dlgOpen.InitialDir );
  finally
    Ini.Free;
  end;
end;

procedure TformMain.cmdExitClick(Sender: TObject);
begin
  Close;
end;

procedure TformMain.lstFilesInsert(Sender: TObject; Item: TListItem);
begin
  mnuEdit.Enabled := true;
  mnuTools.Enabled := true;
  cmdSave.Enabled := true;
end;

procedure TformMain.lstFilesDeletion(Sender: TObject; Item: TListItem);
begin
  if( lstFiles.Items.Count <= 1 ) then begin
    // The last item is being deleted
    mnuEdit.Enabled := false;
    mnuTools.Enabled := false;
    cmdSave.Enabled := false;
  end;
end;

procedure TformMain.lstFilesKeyPress(Sender: TObject; var Key: Char);
begin
  if( Key = #13 ) then lstFilesDblClick( Sender );
end;

function TformMain.AutoSave( Item: TListItem; var Res: integer ): integer;
var
  Info: TId3Info;
begin
  Info := TId3Info( Item.Data );
  if( Info.Changed ) then case Res of
    mrAll: Info.SaveToFile( Item.Caption );
    mrIgnore: ; // No to all
    else begin
      formConfirmSave.edtFilename.Text := ExtractFileName(Item.Caption);
      Res := formConfirmSave.ShowModal;
      if( Res = mrYes )or( Res = mrAll ) then begin
        Info.SaveToFile( Item.Caption );
      end;
    end;
  end;
  Result := Res;
end;

procedure TformMain.ToClipboard( Cut: boolean );
var
  Item: TListItem;
  Data: string;
  Res : integer;
begin
  // Check if some item must be saved
  if( Cut ) then begin
    Res := mrNone;
    Item := lstFiles.Selected;
    while( Item <> nil ) do begin
      if( AutoSave( Item, Res ) = mrCancel ) then
        Exit;
      Item := NextListItem( Item, [isSelected] );
    end
  end;

  lstFiles.Items.BeginUpdate;
  Item := lstFiles.Selected;
  while( Item <> nil ) do begin
    Data := Data + Item.Caption + #13#10;
    if( Cut ) then begin
      TId3Info(Item.Data).Free;
      Item.Delete;
      Item := lstFiles.Selected;
    end
    else begin
      Item := NextListItem( Item, [isSelected] );
    end;
  end;
  lstFiles.Items.EndUpdate;

  if( Data = '' ) then
    Exit;

  // Add to clipboard
  try
    Clipboard.AsText := Data;
  except
  end;
end;

procedure TformMain.cmdCutClick(Sender: TObject);
begin
  ToClipboard( true );
end;

procedure TformMain.cmdCopyClick(Sender: TObject);
begin
  ToClipboard( false );
end;

procedure TformMain.mnuEditClick(Sender: TObject);
begin
  cmdPaste.Enabled := Clipboard.Provides('text/plain');
end;

procedure TformMain.cmdGuessTrackNumberClick(Sender: TObject);
function IsDigit( C: char ): boolean;
  begin
    Result := C in ['0'..'9'];
  end;
function IsLSepar( C: char ): boolean;
  begin
    Result := C in [ ' ', '-', '(', '[' ];
  end;
function IsRSepar( C: char ): boolean;
  begin
    Result := C in [ ' ', '-', ')', ']' ];
  end;
var
  Item: TListItem;
  Info: TId3Info;
  S: string;
  T: string;
  V: string;
  I: integer;
  N: integer;
  L: TStringList;
begin
  L := TStringList.Create;
  try
    Item := lstFiles.Selected;
    while( Item <> nil ) do begin
      S := ' ' + ExtractFileName(Item.Caption);
      // Remove extension
      for I:=Length(S) downto 1 do begin
        if( S[I] = '.' ) then begin
          Delete( S, I, 255 );
          break;
        end;
      end;
      // Parse filename
      L.Clear;
      S := Trim( S ) + '-';
      T := S;
      N := Pos( '-', T );
      while( N > 0 ) do begin
        V := Trim( Copy(T, 1, N-1) );
        if( V <> '' ) then L.AddObject( V, TObject(N) );
        Delete( T, 1, N );
        T := Trim( T );
        N := Pos( '-', T );
      end;
      // Fields are usually ordered in this way: artist, album, track, title
      Info := TId3Info( Item.Data );
      N := L.Count;
      if( Info.Title = '' ) then Info.Title := L.Strings[N-1];
      if( Info.Artist = '' )and( N >= 2 ) then Info.Artist := L.Strings[0];
      if( Info.Album = '' )and( N >= 3 ) then Info.Album := L.Strings[1];
      if( Info.Track = 0 ) then begin
        if( N >= 4 ) then begin
          T := L.Strings[2];
          if( Length(T) = 2 )and IsDigit(T[1]) and IsDigit(T[2]) then
            Info.Track := 10*(Ord(T[1]) - Ord('0')) + Ord(T[2]) - Ord('0')
        end;
        if( Info.Track = 0 ) then begin // Use alternate (aggressive) method
          S := '-' + S;
          for I:=2 to Length(S)-2 do begin
            if( IsLSepar(S[I-1]) and IsDigit(S[I]) and IsDigit(S[I+1]) and IsRSepar(S[I+2]) ) then begin
              N := 10*(Ord(S[I]) - Ord('0')) + Ord(S[I+1]) - Ord('0');
              if( N >= 0 )and( N <= 99 )and( N <> Info.Track ) then begin
                Info.Track := N;
                break;
              end;
            end;
          end;
        end;
        setId3OfItem( Item );
      end;
      Item := NextListItem( Item, [isSelected] );
    end;
  finally
    L.Free;
  end;
end;

(*
  Original function doesn't seem to work well when there is
  only one item in the list and it is selected.
*)
function TformMain.NextListItem( Item: TListItem; States: TItemStates ): TListItem;
begin
  Result := lstFiles.GetNextItem( Item, sdBelow, States );
  if( Result = Item ) then Result := nil;
end;

procedure TformMain.cmdSavePlaylistClick(Sender: TObject);
var
  Item: TListItem;
  Info: TId3Info;
  List: System.Text;
begin
  Item := lstFiles.Selected;
  if( Item = nil ) then
    Exit;
  Info := TId3Info( Item.Data );
  dlgSavePlayList.FileName := Info.Album;
  if( dlgSavePlayList.Execute ) then begin
    try
      AssignFile( List, dlgSavePlayList.FileName );
      ReWrite( List );
      while( Item <> nil ) do begin
        WriteLn( List, ExtractFileName(Item.Caption) );
        Item := NextListItem( Item, [isSelected] );
      end;
      CloseFile( List );
    except
      Application.MessageBox( 'There was an error saving the playlist.', 'Error', [smbOk], smsCritical );
    end;
  end;
end;

procedure TformMain.cmdMakeHTMLIndexClick(Sender: TObject);
var
  Item: TListItem;
  Info: TId3Info;
  Inf1: TId3Info;
  Line: string;
  Rows: integer;
  Cols: integer;
  Temp: string;
  Name: string;
  InpFile: System.Text;
  OutFile: System.Text;
  I: integer;
begin
  Item := lstFiles.Selected;
  if( Item = nil ) then
    Exit;
  if not( formIndexProperties.ShowModal = mrOk ) then
    Exit;
  if not( dlgSaveIndex.Execute ) then
    Exit;
  Cols := formIndexProperties.lstFields.Items.Count;
  Rows := lstFiles.SelCount;
  try
    AssignFile( OutFile, dlgSaveIndex.FileName );
    AssignFile( InpFile, formIndexProperties.edtHeader.Text );
    Reset( InpFile );
    ReWrite( OutFile );
    Inf1 := TId3Info( Item.Data );
    while( not Eof(InpFile) ) do begin
      if( Line = sExpandTable ) then begin
        while( Item <> nil ) do begin
          Info := TId3Info( Item.Data );
          Name := ExtractFileName( Item.Caption );
          Line := '<TR>';
          for I:=0 to Cols-1 do begin
            Temp := formIndexProperties.lstFields.Items[I];
            Delete( Temp, 1, Pos(#9,Temp) );
            Line := Line + '<TD>' + Temp + '</TD>';
          end;
          Line := Line + '</TR>';
          ReplaceId3Tokens( Line, Info );
          StringReplace( Line, sFilePath, Item.Caption );
          StringReplace( Line, sFileName, Name );
          WriteLn( OutFile, Line );
          Item := NextListItem( Item, [isSelected] );
        end;
      end;
      ReadLn( InpFile, Line );
      ReplaceId3Tokens( Line, Inf1 );
      StringReplace( Line, sTabCols, IntToStr(Cols) );
      StringReplace( Line, sTabRows, IntToStr(Rows) );
      WriteLn( OutFile, Line );
    end;
    CloseFile( InpFile );
    CloseFile( OutFile );
  except
  end;
end;

procedure TformMain.cmdCapitalizeTitleClick(Sender: TObject);
var
  Item: TListItem;
  Info: TId3Info;
  Name: string;
  I: integer;
begin
  lstFiles.Items.BeginUpdate;
  Item := lstFiles.Selected;
  while( Item <> nil ) do begin
    Info := TId3Info( Item.Data );
    Name := Info.Title;
    I := 1;
    while( I > 0 )and( I < Length(Name) ) do begin
      Name[I] := UpCase( Name[I] );
      while( I < Length(Name) )and( Name[I] <> ' ' ) do
        Inc( I );
      while( I < Length(Name) )and( Name[I] = ' ' ) do
        Inc( I );
    end;
    if( Name <> Info.Title ) then begin
      Info.Title := Name;
      setId3OfItem( Item );
    end;
    Item := NextListItem( Item, [isSelected] );
  end;
  lstFiles.Items.EndUpdate;
end;

procedure TformMain.tbbNewClick(Sender: TObject);
var
  Idx: integer;
begin
  if( not VerifySaved( false ) ) then
    Exit;

  for Idx:=0 to lstFiles.Items.Count-1 do
    TId3Info(lstFiles.Items[Idx].Data).Free;
  lstFiles.Items.Clear;
end;

procedure TformMain.lstFilesEdited(Sender: TObject; Item: TListItem;
  var S: WideString);
begin
  if( not RenameFile( Item.Caption, S ) ) then begin
    Application.MessageBox( 'Cannot rename the file.', 'Error', [smbOk], smsCritical );
    S := Item.Caption;
  end;
end;

end.
